# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from datetime import datetime, timedelta
from typing import Optional

from spectre_core.spectrograms.array_operations import find_closest_index
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.cfg import DEFAULT_DATETIME_FORMAT


def _average_array(array: np.ndarray, average_over: int, axis=0) -> np.ndarray:

    # Check if average_over is an integer
    if type(average_over) != int:
        raise TypeError(f"average_over must be an integer. Got {type(average_over)}")

    # Get the size of the specified axis which we will average over
    axis_size = array.shape[axis]
    # Check if average_over is within the valid range
    if not 1 <= average_over <= axis_size:
        raise ValueError(f"average_over must be between 1 and the length of the axis ({axis_size})")
    
    max_axis_index = len(np.shape(array)) - 1
    if axis > max_axis_index: # zero indexing on specifying axis, so minus one
        raise ValueError(f"Requested axis is out of range of array dimensions. Axis: {axis}, max axis index: {max_axis_index}")

    # find the number of elements in the requested axis
    num_elements = array.shape[axis]

    # find the number of "full blocks" to average over
    num_full_blocks = num_elements // average_over
    # if num_elements is not exactly divisible by average_over, we will have some elements left over
    # these remaining elements will be padded with nans to become another full block
    remainder = num_elements % average_over
    
    # if there exists a remainder, pad the last block
    if remainder != 0:
        # initialise an array to hold the padding shape
        padding_shape = [(0, 0)] * array.ndim
        # pad after the last column in the requested axis
        padding_shape[axis] = (0, average_over - remainder)
        # pad with nan values (so to not contribute towards the mean computation)
        array = np.pad(array, padding_shape, mode='constant', constant_values=np.nan)
    
    # initalise a list to hold the new shape
    new_shape = list(array.shape)
    # update the shape on the requested access (to the number of blocks we will average over)
    new_shape[axis] = num_full_blocks + (1 if remainder else 0)
    # insert a new dimension, with the size of each block
    new_shape.insert(axis + 1, average_over)
    # and reshape the array to sort the array into the relevant blocks.
    reshaped_array = array.reshape(new_shape)
    # average over the newly created axis, essentially averaging over the blocks.
    averaged_array = np.nanmean(reshaped_array, axis=axis + 1)
    # return the averaged array
    return averaged_array


def frequency_chop(input_spectrogram: Spectrogram, 
                   start_frequency: float | int, 
                   end_frequency: float | int) -> Optional[Spectrogram]:
    
    is_entirely_below_frequency_range = (start_frequency < input_spectrogram.frequencies[0] and end_frequency < input_spectrogram.frequencies[0])
    is_entirely_above_frequency_range = (start_frequency > input_spectrogram.frequencies[-1] and end_frequency > input_spectrogram.frequencies[-1])
    # if the requested frequency range is out of bounds for the spectrogram return None
    if is_entirely_below_frequency_range or is_entirely_above_frequency_range:
        return None
    
    #find the index of the nearest matching frequency bins in the spectrogram
    start_index = find_closest_index(start_frequency, input_spectrogram.frequencies)
    end_index = find_closest_index(end_frequency, input_spectrogram.frequencies)
    
    # enforce distinct start and end indices
    if start_index == end_index:
        raise ValueError(f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}")  
    
    # if start index is more than end index, swap the ordering so to enforce start_index <= end_index
    if start_index > end_index:
        start_index, end_index = end_index, start_index
    
    # chop the spectrogram accordingly
    transformed_dynamic_spectra = input_spectrogram.dynamic_spectra[start_index:end_index+1, :]
    transformed_frequencies = input_spectrogram.frequencies[start_index:end_index+1]
    
    # return the spectrogram instance
    return Spectrogram(transformed_dynamic_spectra,
                       input_spectrogram.times,
                       transformed_frequencies,
                       input_spectrogram.tag,
                       chunk_start_time = input_spectrogram.chunk_start_time,
                       microsecond_correction = input_spectrogram.microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def time_chop(input_spectrogram: Spectrogram, 
              start_time: str, 
              end_time: str, 
              time_format: str = DEFAULT_DATETIME_FORMAT) -> Optional[Spectrogram]:
    
    # parse the strings as datetimes
    start_datetime = datetime.strptime(start_time, time_format)
    end_datetime = datetime.strptime(end_time, time_format)

    # if the requested time range is out of bounds for the spectrogram return None
    is_entirely_below_time_range = (start_datetime < input_spectrogram.datetimes[0] and end_datetime < input_spectrogram.datetimes[0])
    is_entirely_above_time_range = (start_datetime > input_spectrogram.datetimes[-1] and end_datetime > input_spectrogram.datetimes[-1])
    if is_entirely_below_time_range or is_entirely_above_time_range:
        return None
    
    start_index = find_closest_index(start_datetime, input_spectrogram.datetimes)
    end_index = find_closest_index(end_datetime, input_spectrogram.datetimes)
    
    if start_index == end_index:
        raise ValueError(f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}")
    
    if start_index > end_index:
        start_index, end_index = end_index, start_index

    # chop the spectrogram 
    transformed_dynamic_spectra = input_spectrogram.dynamic_spectra[:, start_index:end_index+1]

    # chop the times array
    transformed_times = input_spectrogram.times[start_index:end_index+1]
    #translate the chopped times array to start at zero
    transformed_times -= transformed_times[0]

    # compute the new start datetime following the time chop
    transformed_start_datetime = input_spectrogram.datetimes[start_index]
    # parse the chunk start time (as string)
    transformed_chunk_start_time = datetime.strftime(transformed_start_datetime, DEFAULT_DATETIME_FORMAT)
    # and compute the microsecond correction
    transformed_microsecond_correction = transformed_start_datetime.microsecond

    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       input_spectrogram.frequencies, 
                       input_spectrogram.tag, 
                       chunk_start_time = transformed_chunk_start_time,
                       microsecond_correction = transformed_microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def time_average(input_spectrogram: Spectrogram, 
                 average_over: int) -> Spectrogram:
    
    # spectre does not currently support averaging of non-datetime assigned spectrograms
    if input_spectrogram.chunk_start_time is None:
        raise ValueError(f"Input spectrogram is missing chunk start time. Averaging is not yet supported for non-datetime assigned spectrograms")
    
    # if the user has requested no averaging, just return the same instance unchanged
    if average_over == 1:
        return input_spectrogram

    # average the dynamic spectra array
    transformed_dynamic_spectra = _average_array(input_spectrogram.dynamic_spectra, average_over, axis=1)
    # average the times array s.t. the ith averaged spectrum is assigned the 
    transformed_times = _average_array(input_spectrogram.times, average_over)

    # We need to assign timestamps to the averaged spectrums in the spectrograms. The natural way to do this
    # is to assign the i'th averaged spectrogram to the i'th averaged time stamp. From this,
    # we then need to compute the chunk start time to assig to the first averaged spectrum,
    # and update the microsecond correction.
    
    # define the initial spectrum as the spectrum at time index 0 in the spectrogram
    # then, averaged_t0 is the seconds elapsed between the input intial spectrum and the averaged intial spectrum
    averaged_t0 = float(transformed_times[0])
    # compute the updated chunk start time and the updated microsecond correction based on averaged_t0
    corrected_start_datetime = input_spectrogram.datetimes[0] + timedelta(seconds = averaged_t0)
    # then, compute the transformed chunk start time and microsecond correction
    transformed_chunk_start_time = corrected_start_datetime.strftime(DEFAULT_DATETIME_FORMAT)
    transformed_microsecond_correction = corrected_start_datetime.microsecond

    # finally, translate the averaged time seconds to begin at t=0 [s]
    transformed_times -= transformed_times[0]
    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       input_spectrogram.frequencies, 
                       input_spectrogram.tag,
                       chunk_start_time = transformed_chunk_start_time,
                       microsecond_correction = transformed_microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)

def frequency_average(input_spectrogram: Spectrogram, 
                      average_over: int) -> Spectrogram:
    
    # if the user has requested no averaging, just return the same instance unchanged
    if average_over == 1:
        return input_spectrogram

    # We need to assign physical frequencies to the averaged spectrums in the spectrograms.
    # is to assign the i'th averaged spectral component to the i'th averaged frequency.
    # average the dynamic spectra array
    transformed_dynamic_spectra = _average_array(input_spectrogram.dynamic_spectra, average_over, axis=0)
    transformed_frequencies = _average_array(input_spectrogram.frequencies, average_over)

    return Spectrogram(transformed_dynamic_spectra, 
                       input_spectrogram.times, 
                       transformed_frequencies, 
                       input_spectrogram.tag,
                       chunk_start_time = input_spectrogram.chunk_start_time, 
                       microsecond_correction = input_spectrogram.microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def _time_elapsed(datetimes: np.ndarray) -> np.ndarray:
    # Extract the first datetime to use as the reference point
    base_time = datetimes[0]
    # Calculate elapsed time in seconds for each datetime in the list
    elapsed_time = [(dt - base_time).total_seconds() for dt in datetimes]
    # Convert the list of seconds to a NumPy array of type float32
    return np.array(elapsed_time, dtype=np.float32)

# we assume that the spectrogram list is ORDERED chronologically
# we assume there is no time overlap in any of the spectrograms in the list
def join_spectrograms(spectrograms: list[Spectrogram]) -> Spectrogram:

    # check that the length of the list is non-zero
    num_spectrograms = len(spectrograms)
    if num_spectrograms == 0:
        raise ValueError(f"Input list of spectrograms is empty!")
    
    # extract the first element of the list, and use this as a reference for comparison
    # input validations.
    reference_spectrogram = spectrograms[0] 

    # perform checks on each spectrogram in teh list
    for spectrogram in spectrograms:
        if not np.all(np.equal(spectrogram.frequencies, reference_spectrogram.frequencies)):
            raise ValueError(f"All spectrograms must have identical frequency ranges")
        if spectrogram.tag != reference_spectrogram.tag:
            raise ValueError(f"All tags must be equal for each spectrogram in the input list!")
        if spectrogram.spectrum_type != reference_spectrogram.spectrum_type:
            raise ValueError(f"All units must be equal for each spectrogram in the input list!")
        if spectrogram.chunk_start_time is None:
            raise ValueError(f"All spectrograms must have chunk_start_time set. Received {spectrogram.chunk_start_time}")


    # build a list of the time array of each spectrogram in the list
    conc_datetimes = np.concatenate([s.datetimes for s in spectrograms])
    # find the total number of time stamps
    num_total_time_bins = len(conc_datetimes)
    # find the total number of frequency bins (we can safely now use the first)
    num_total_freq_bins = len(reference_spectrogram.frequencies)
    # create an empty numpy array to hold the joined spectrograms
    transformed_dynamic_spectra = np.empty((num_total_freq_bins, num_total_time_bins))

    start_index = 0
    for spectrogram in spectrograms:
        end_index = start_index + len(spectrogram.times)
        transformed_dynamic_spectra[:, start_index:end_index] = spectrogram.dynamic_spectra
        start_index = end_index

    transformed_times = _time_elapsed(conc_datetimes)
    
    transformed_microsecond_correction = conc_datetimes[0].microsecond

    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       reference_spectrogram.frequencies, 
                       reference_spectrogram.tag, 
                       chunk_start_time = reference_spectrogram.chunk_start_time,
                       microsecond_correction = transformed_microsecond_correction,
                       spectrum_type = reference_spectrogram.spectrum_type) 




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.file_handlers.base import BaseFileHandler

class TextHandler(BaseFileHandler):
    def __init__(self, 
                 parent_path: str,
                 base_file_name: str, 
                 extension: str = "txt",
                 **kwargs):
        super().__init__(parent_path,
                         base_file_name,
                         extension, 
                         **kwargs)
    

    def read(self) -> dict:
        with open(self.file_path, 'r') as f:
            return f.read()


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional

from spectre_core.spectrograms.spectrogram import TimeCut
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.plotting.base import BaseTimeSeriesPanel, CutsPanel
from spectre_core.plotting.format import DEFAULT_FORMATS
from spectre_core.plotting.panel_register import register_panel

TIME_CUTS_PANEL_NAME = "time_cuts"

@register_panel(TIME_CUTS_PANEL_NAME)
class Panel(BaseTimeSeriesPanel, CutsPanel):
    def __init__(self, 
                 name: str,
                 spectrogram: Spectrogram, 
                 time_type: str = "seconds",
                 *frequencies: list[float],
                 dBb: bool = False,
                 peak_normalise: bool = False,
                 background_subtract: bool = False,
                 cmap: str = DEFAULT_FORMATS.cuts_cmap,
                 **kwargs):
        super().__init__(name, 
                         spectrogram, 
                         time_type, 
                         **kwargs)
        self._frequencies = frequencies
        self._cmap = cmap
        self._dBb = dBb
        self._peak_normalise = peak_normalise
        self._background_subtract = background_subtract
        # map each cut frequency to the corresponding TimeCut dataclass
        self._time_cuts: Optional[dict[float, TimeCut]] = {} 
    

    @property
    def time_cuts(self) -> dict[float, TimeCut]:
        if not self._time_cuts:
            for frequency in self._frequencies:
                time_cut =  self._spectrogram.get_time_cut(frequency,
                                                           dBb = self._dBb,
                                                           peak_normalise = self._peak_normalise,
                                                           correct_background = self._background_subtract,
                                                           return_time_type=self._time_type)
                self._time_cuts[time_cut.frequency] = time_cut
        return self._time_cuts
    

    @property
    def frequencies(self) -> list[float]:
        return list(self.time_cuts.keys())


    def draw(self):
        for frequency, color in self.bind_to_colors():
            time_cut = self.time_cuts[frequency]
            self.ax.step(self.times, 
                         time_cut.cut, 
                         where='mid', 
                         color = color)
    

    def annotate_y_axis(self) -> None:
        if self._dBb:
            self.ax.set_ylabel('dBb')
        elif self._peak_normalise:
            return # no y-axis label
        else:
            self.ax.set_ylabel(f'{self._spectrogram.spectrum_type.capitalize()}')

    
    def bind_to_colors(self):
        return super().bind_to_colors(self.frequencies, cmap = self._cmap)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from abc import ABC, abstractmethod
from typing import Any, Optional
from warnings import warn

class BaseFileHandler(ABC):
    def __init__(self, 
                 parent_path: str, 
                 base_file_name: str, 
                 extension: Optional[str] = None):
        self._parent_path = parent_path
        self._base_file_name = base_file_name
        self._extension = extension

        
    @abstractmethod
    def read(self) -> Any:
        pass
 

    @property
    def parent_path(self) -> str:
        return self._parent_path
    

    @property
    def base_file_name(self) -> str:
        return self._base_file_name
    

    @property
    def extension(self) -> Optional[str]:
        return self._extension
    

    @property
    def file_name(self) -> str:
        return self._base_file_name if (self._extension is None) else f"{self._base_file_name}.{self._extension}"
    

    @property
    def file_path(self) -> str:
        return os.path.join(self._parent_path, self.file_name)
    
    
    @property
    def exists(self) -> bool:
        return os.path.exists(self.file_path) 


    def make_parent_path(self) -> None:
        os.makedirs(self.parent_path, exist_ok=True) 
    

    def delete(self,
               ignore_if_missing: bool = False) -> None:
        if not self.exists and not ignore_if_missing:
            raise FileNotFoundError(f"{self.file_name} does not exist, and so cannot be deleted")
        else:
            os.remove(self.file_path)
    

    def cat(self) -> None:
        print(self.read())


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from importlib import import_module

def import_target_modules(caller_file: str, # __file__ in the calling context for the library
                          caller_name: str, # __name__ in the calling context for the library
                          target_module: str # the module we are looking to dynamically import
) -> None: 
    # fetch the directory path for the __init__.py in the library directory
    library_dir_path = os.path.dirname(caller_file)
    # list all subdirectories in the library directory
    subdirs = [x.name for x in os.scandir(library_dir_path) if x.is_dir() and (x.name != "__pycache__")]
    # for each subdirectory, try and import the target module
    for subdir in subdirs:
        full_module_name = f"{caller_name}.{subdir}.{target_module}"
        import_module(full_module_name)






# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# dynamically import all receiver
import spectre_core.receivers.library


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import os
import time
from queue import Queue
from typing import Any
from abc import ABC, abstractmethod
from math import floor

from watchdog.events import FileSystemEventHandler

from spectre_core.chunks.factory import get_chunk_from_tag
from spectre_core.file_handlers.configs import CaptureConfig
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.spectrograms.transform import join_spectrograms
from spectre_core.spectrograms.transform import (
    time_average, 
    frequency_average
)


class BaseEventHandler(ABC, FileSystemEventHandler):
    def __init__(self, 
                 tag: str, 
                 exception_queue: Queue, 
                 extension: str):
        self._tag = tag
        self._Chunk = get_chunk_from_tag(tag)

        self._capture_config = CaptureConfig(tag)

        self._extension = extension
        self._exception_queue = exception_queue  # Queue to propagate exceptions

        self._spectrogram: Spectrogram = None # spectrogram cache


    @abstractmethod
    def process(self, file_path: str) -> None:
        pass


    def on_created(self, event):
        if not event.is_directory and event.src_path.endswith(self._extension):
            _LOGGER.info(f"Noticed: {event.src_path}")
            try:
                self._wait_until_stable(event.src_path)
                self.process(event.src_path)
            except Exception as e:
                _LOGGER.error(f"An error has occured while processing {event.src_path}",
                              exc_info=True)
                self._flush_spectrogram() # flush the internally stored spectrogram
                # Capture the exception and propagate it through the queue
                self._exception_queue.put(e)


    def _wait_until_stable(self, file_path: str):
        _LOGGER.info(f"Waiting for file stability: {file_path}")
        size = -1
        while True:
            current_size = os.path.getsize(file_path)
            if current_size == size:
                _LOGGER.info(f"File is now stable: {file_path}")
                break  # File is stable when the size hasn't changed
            size = current_size
            time.sleep(0.25)


    def _average_in_time(self, spectrogram: Spectrogram) -> Spectrogram:
        requested_time_resolution = self._capture_config.get('time_resolution') # [s]
        if requested_time_resolution is None:
            raise KeyError(f"Time resolution has not been specified in the capture config!")
        average_over = floor(requested_time_resolution/spectrogram.time_resolution) if requested_time_resolution > spectrogram.time_resolution else 1
        return time_average(spectrogram, average_over)
    
    
    def _average_in_frequency(self, spectrogram: Spectrogram) -> Spectrogram:
        frequency_resolution = self._capture_config.get('frequency_resolution') # [Hz]
        if frequency_resolution is None:
            raise KeyError(f"Frequency resolution has not been specified in the capture config!")
        average_over = floor(frequency_resolution/spectrogram.frequency_resolution) if frequency_resolution > spectrogram.frequency_resolution else 1
        return frequency_average(spectrogram, average_over)
    

    def _join_spectrogram(self, spectrogram: Spectrogram) -> None:
        if self._spectrogram is None:
            self._spectrogram = spectrogram
        else:
            self._spectrogram = join_spectrograms([self._spectrogram, spectrogram])

        if self._spectrogram.time_range >= self._capture_config.get("joining_time"):
            self._flush_spectrogram()
    

    def _flush_spectrogram(self) -> None:
        if self._spectrogram:
            _LOGGER.info(f"Flushing spectrogram to file with chunk start time {self._spectrogram.chunk_start_time}")
            self._spectrogram.save()
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self._spectrogram = None # reset the cache


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta
from typing import Tuple
import numpy as np

from astropy.io import fits
from astropy.io.fits.hdu.image import PrimaryHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.hdu.hdulist import HDUList

from spectre_core.chunks.chunk_register import register_chunk
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.chunks.base import (
    SPECTREChunk, 
    ChunkFile
)

@register_chunk("fixed")
class Chunk(SPECTREChunk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs) 
        
        self.add_file(BinChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(FitsChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(HdrChunk(self.chunk_parent_path, self.chunk_name))


    def build_spectrogram(self) -> Spectrogram:
        """Create a spectrogram by performing a Short Time FFT on the IQ samples for this chunk."""
        iq_data = self.read_file("bin")
        millisecond_correction = self.read_file("hdr")

        # units conversion
        microsecond_correction = millisecond_correction * 1e3

        times, frequencies, dynamic_spectra = self.__do_STFFT(iq_data)

        # explicitly type cast data arrays to 32-bit floats
        times = np.array(times, dtype = 'float32')
        frequencies = np.array(frequencies, dtype = 'float32')
        dynamic_spectra = np.array(dynamic_spectra, dtype = 'float32')

        return Spectrogram(dynamic_spectra, 
                           times, 
                           frequencies, 
                           self.tag, 
                           chunk_start_time = self.chunk_start_time, 
                           microsecond_correction = microsecond_correction,
                           spectrum_type = "amplitude")


    def __do_STFFT(self, 
                   iq_data: np.array) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """For reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.ShortTimeFFT.html"""

        # set p0=0, since by convention in the STFFT docs, p=0 corresponds to the slice centred at t=0
        p0=0

        # set p1 to the index of the first slice where the "midpoint" of the window is still inside the signal
        num_samples = len(iq_data)
        p1 = self.SFT.upper_border_begin(num_samples)[1]
        
        # compute a ShortTimeFFT on the IQ samples
        complex_spectra = self.SFT.stft(iq_data, 
                                        p0 = p0, 
                                        p1 = p1) 
        
        # compute the magnitude of each spectral component
        dynamic_spectra = np.abs(complex_spectra)


        # assign a physical time to each spectrum
        times = self.SFT.t(num_samples, 
                           p0 = 0, 
                           p1 = p1)

        # assign physical frequencies to each spectral component
        frequencies = self.SFT.f + self.capture_config.get('center_freq', 0.0) 

        return times, frequencies, dynamic_spectra


class BinChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "bin")

    def read(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.complex64)


  
class HdrChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "hdr")


    def read(self) -> int:
        hdr_contents = self._extract_contents()
        return self._get_millisecond_correction(hdr_contents)


    def _extract_contents(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.float32)


    def _get_millisecond_correction(self, hdr_contents: np.ndarray) -> int:
        if len(hdr_contents) != 1:
            raise ValueError(f"Expected exactly one integer in the header, but received header contents: {hdr_contents}")
        
        millisecond_correction_as_float = float(hdr_contents[0])

        if not millisecond_correction_as_float.is_integer():
            raise TypeError(f"Expected integer value for millisecond correction, but got {millisecond_correction_as_float}")
        
        return int(millisecond_correction_as_float)
        



class FitsChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "fits")


    @property
    def datetimes(self) -> np.ndarray:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            bintable_data = hdulist[1].data
            times = bintable_data['TIME'][0]
            return [self.chunk_start_datetime + timedelta(seconds=t) for t in times]
        

    def read(self) -> Spectrogram:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            primary_hdu = self._get_primary_hdu(hdulist)
            dynamic_spectra = self._get_dynamic_spectra(primary_hdu)
            spectrum_type = self._get_spectrum_type(primary_hdu)
            microsecond_correction = self._get_microsecond_correction(primary_hdu)
            bintable_hdu = self._get_bintable_hdu(hdulist)
            times, frequencies = self._get_time_and_frequency(bintable_hdu)

        return Spectrogram(dynamic_spectra, 
                           times, 
                           frequencies, 
                           self.tag, 
                           chunk_start_time=self.chunk_start_time, 
                           microsecond_correction=microsecond_correction,
                           spectrum_type = spectrum_type)


    def _get_primary_hdu(self, hdulist: HDUList) -> PrimaryHDU:
        return hdulist['PRIMARY']


    def _get_dynamic_spectra(self, primary_hdu: PrimaryHDU) -> np.ndarray:
        return primary_hdu.data


    def _get_spectrum_type(self, primary_hdu: PrimaryHDU) -> str:
        return primary_hdu.header.get('BUNIT', None)


    def _get_microsecond_correction(self, primary_hdu: PrimaryHDU) -> int:
        date_obs = primary_hdu.header.get('DATE-OBS', None)
        time_obs = primary_hdu.header.get('TIME-OBS', None)
        datetime_obs = datetime.strptime(f"{date_obs}T{time_obs}", "%Y-%m-%dT%H:%M:%S.%f")
        return datetime_obs.microsecond


    def _get_bintable_hdu(self, hdulist: HDUList) -> BinTableHDU:
        return hdulist[1]


    def _get_time_and_frequency(self, bintable_hdu: BinTableHDU) -> Tuple[np.ndarray, np.ndarray]:
        data = bintable_hdu.data
        times = data['TIME'][0]
        frequencies_MHz = data['FREQUENCY'][0]
        frequencies = frequencies_MHz * 1e6 # convert to Hz
        return times, frequencies




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional
from datetime import datetime

from spectre_core.spectrograms.spectrogram import FrequencyCut
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.plotting.base import BaseSpectrumPanel, CutsPanel
from spectre_core.plotting.panel_register import register_panel
from spectre_core.plotting.format import DEFAULT_FORMATS

FREQUENCY_CUTS_PANEL_NAME = "frequency_cuts"

@register_panel(FREQUENCY_CUTS_PANEL_NAME)
class Panel(BaseSpectrumPanel, CutsPanel):
    def __init__(self, 
                 name: str,
                 spectrogram: Spectrogram, 
                 time_type: str = "seconds",
                 *times: list[float | str],
                 dBb: bool = False,
                 peak_normalise: bool = False,
                 cmap: str = DEFAULT_FORMATS.cuts_cmap,
                 **kwargs):
        super().__init__(name, 
                         spectrogram, 
                         time_type, 
                         **kwargs)
        self._times = times
        self._cmap = cmap
        self._dBb = dBb
        self._peak_normalise = peak_normalise
        # map each time cut to the corresponding FrequencyCut dataclass
        self._frequency_cuts: Optional[dict[float | datetime, FrequencyCut]] = {}


    @property
    def frequency_cuts(self) -> dict[float | str, FrequencyCut]:
        if not self._frequency_cuts:
            for time in self._times:
                frequency_cut = self._spectrogram.get_frequency_cut(time,
                                                                    dBb = self._dBb,
                                                                    peak_normalise = self._peak_normalise)
                self._frequency_cuts[frequency_cut.time] = frequency_cut
        return self._frequency_cuts


    @property
    def times(self) -> list[float | datetime]:
        return list(self.frequency_cuts.keys())
    

    def draw(self):
        for time, color in self.bind_to_colors():
            frequency_cut = self.frequency_cuts[time]
            self.ax.step(self.frequencies*1e-6, # convert to MHz
                         frequency_cut.cut, 
                         where='mid', 
                         color = color)
    

    def annotate_y_axis(self) -> None:
        if self._dBb:
            self.ax.set_ylabel('dBb')
        elif self._peak_normalise:
            return # no y-axis label
        else:
            self.ax.set_ylabel(f'{self._spectrogram.spectrum_type.capitalize()}')

    
    def bind_to_colors(self):
        return super().bind_to_colors(self.times, cmap = self._cmap)